using System;
using UnityEngine;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using MVR.FileManagementSecure;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

namespace Feel
{
    public class DildoMovement : MVRScript
    {
        #region PluginInfo    
        public string pluginAuthor = "Feel";
        public string pluginName = "DildoMovement";
        public string pluginVersion = "1";
        public string pluginDate = "02/10" + "/2023";
        public string pluginDescription = @"Dildo Movement Plugin";
        #endregion

        #region Vars    

        protected bool first_loading = false;

        private Atom _thisAtom;
        private FreeControllerV3 _thisControl;
        public Transform ObjectRoot;
        private Atom _selectAtom;

        private JSONStorableBool IsMovie;
        private JSONStorableFloat VibrationWidth;
        private JSONStorableFloat IntervalTime;
        private JSONStorableFloat _DirectionX;
        private JSONStorableFloat _DirectionY;
        private JSONStorableFloat _DirectionZ;

        private bool editorInitialized = false;
        private int initLoops = 0;
        private int maxInitLoops = 1500; // 3600 for one minute
        
        private bool _is_in_out = false;
        private Vector3 _base_vector;

        private float _move_value = 0f;
        private float _move_width = 0f;
        private float _move_gap = 0f;
        private float _move_time = 1f;

        private EventTrigger _movie_start_event;
        private EventTrigger _movie_end_event;

        #endregion

        public override void Init()
        {
            try
            {
                if (containingAtom.type != "CustomUnityAsset" && containingAtom.type != "Capsule" && containingAtom.type != "Dildo" && containingAtom.type != "Paddle"
                     && containingAtom.type != "ToyAH" && containingAtom.type != "ToyBP")
                {
                    SuperController.LogError("DildoMovement - An object of the wrong type. ( CustomUnityAsset, Capsule, Dildo, Paddle, ToyAH, ToyBP).");
                    return;
                }

                DMEventHandler.LoadAssets();

                _thisAtom = containingAtom;
                _thisControl = _thisAtom.freeControllers.First(fc => fc.name == "control");

                IsMovie = new JSONStorableBool("Dildo Movie", false, OnDildoMovie);
                RegisterBool(IsMovie);
                UIDynamicToggle toggle = CreateToggle(IsMovie, false);
                toggle.label = "Movie On Off";
                toggle.height = 80;

                VibrationWidth = new JSONStorableFloat("vibration width", 0f, OnVibrationWidthChanged, 0f, 1f);
                RegisterFloat(VibrationWidth);
                CreateSlider(VibrationWidth, false);

                IntervalTime = new JSONStorableFloat("interval time", 1f, OnIntervalTimeChanged, 1f, 10f);
                RegisterFloat(IntervalTime);
                CreateSlider(IntervalTime, false);

                _DirectionX = new JSONStorableFloat("direction x", 0f, OnDirectionChanged, 0f, 1f);
                RegisterFloat(_DirectionX);
                CreateSlider(_DirectionX, false);

                _DirectionY = new JSONStorableFloat("direction y", 0f, OnDirectionChanged, 0f, 1f);
                RegisterFloat(_DirectionY);
                CreateSlider(_DirectionY, false);

                _DirectionZ = new JSONStorableFloat("direction z", 1f, OnDirectionChanged, 0f, 1f);
                RegisterFloat(_DirectionZ);
                CreateSlider(_DirectionZ, false);

                //CreateButton("Edit Trigger", OnMovieStartEventFunc, true);
                UIDynamicButton btn1 = CreateButton("Movie Start Event", true);
                btn1.height = 80f;
                btn1.button.onClick.AddListener(OnMovieStartEventFunc);

                UIDynamicButton btn2 = CreateButton("Movie End Event", true);
                btn2.height = 80f;
                btn2.button.onClick.AddListener(OnMovieEndEventFunc);


                // A tiny bit of help to give some info to your fellow creator and/or player
                JSONStorableString helpText = new JSONStorableString("About",
                    "<color=#000><size=35><b>Dildo Movement</b></size></color>\n\n" +
                    "<color=#333>" +
                    "Move or vibrate the asset.\n\n" +
                    "thank you for your support\n" +
                    "<b>https://www.patreon.com/feel_vam</b>\n\n" +
                    "</color>");
                UIDynamic helpTextfield = CreateTextField(helpText, true);
                helpTextfield.height = 200.0f;

                var btn = CreateButton("Go To Support", true);
                btn.button.onClick.AddListener(() =>
                {
                    Application.OpenURL("https://www.patreon.com/feel_vam");
                });

            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        // Update is called with each rendered frame by Unity
        void Update()
        {
            // I don't know why the fuck this asset does not get the children properly during the loading process, so I'm checking the singleton
            // In reality, that's even cleaner to do it that way
            if (initLoops < maxInitLoops && editorInitialized == false && !SuperController.singleton.isLoading)
            {
                initLoops++;
                InitEditor();
            }

            if (initLoops == maxInitLoops)
            {
                SuperController.LogError("DildoMovement : This CUA could not be initialized properly. Please select a compatible object and reload the script.");
                initLoops++; // To avoid spamming the log with the error
            }

            // first set.
            if(first_loading == false)
            {
                first_loading = true;
            }
            
            
        }

        void FixedUpdate()
        {
            if(IsMovie.val && _thisAtom != null)
            {
                if (_is_in_out)
                {
                    //_is_in_out = false;
                    _move_value -= _move_gap;
                    //offsetPosition = new Vector3(0f, 0f, -0.02f * VibrationWidth.val);
                }
                else
                {
                    //_is_in_out = true;
                    _move_value += _move_gap;
                    //offsetPosition = new Vector3(0f, 0f, 0.02f * VibrationWidth.val);
                }

                if (_move_value < (_move_width * -1f))
                {
                    _move_value = _move_width * -1f;
                    _is_in_out = false;
                }
                if (_move_value > _move_width)
                {
                    _move_value = _move_width;
                    _is_in_out = true;
                }

                Vector3 offsetPosition = new Vector3(_move_value * _DirectionX.val, _move_value * _DirectionY.val, _move_value * _DirectionZ.val);

                _thisControl.transform.Translate(offsetPosition, _thisControl.transform);


            }
        }

        void OnDestroy()
        {
            
        }

        private bool InitEditor()
        {
            if (editorInitialized == true) return true;

            // I found my transforms, now the rest
            if (_thisAtom != null)
            {
                editorInitialized = true;

                // Everything is initialized, triggering the callbacks to init the door rotation saved in the scene file
                // I'm just quickly fixing the script for VamTastic, I'm not gonna refactor everything, so in reality,
                // the float val param of the callbacks is useless

                return true;
            }
            else
            {
                SuperController.LogError("DildoMovement : your is missing mandatory object.");
                initLoops = maxInitLoops; // We don't have what we need, we consider the initialization failed
                return false;
            }

            return false;
        }

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
            if (includePhysical || forceStore)
            {
                needsStore = true;
                if (_movie_start_event != null) jc[_movie_start_event.Name] = _movie_start_event.GetJSON(base.subScenePrefix);
                if (_movie_end_event != null) jc[_movie_end_event.Name] = _movie_end_event.GetJSON(base.subScenePrefix);
                //SuperController.LogMessage(">> 1 GetJSON " + _movie_start_event.Name + " - " + base.subScenePrefix);
            }
            //SuperController.LogMessage(">> 1 GetJSON " + jc.ToString());
            return jc;
        }

        public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
        {
            base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);

            //_movie_start_event = new EventTrigger(this, "Movie Start Event");
            if(jc.HasKey("Movie Start Event")) _movie_start_event = new EventTrigger(this, "Movie Start Event");
            if (jc.HasKey("Movie End Event")) _movie_end_event = new EventTrigger(this, "Movie End Event");

            //SuperController.LogMessage(">> 1 RestoreFromJSON " + jc.ToString());

            base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
        }

        public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
        {
            base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);

            if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
            {
                if (jc.HasKey("Movie Start Event"))  _movie_start_event.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
                if (jc.HasKey("Movie End Event")) _movie_end_event.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
            }
        }


        private void OnDildoMovie(JSONStorableBool jb)
        {
            if(_thisControl != null)
            {
                if (jb.val)
                {
                    _is_in_out = false;
                    _move_value = 0f;
                    _base_vector = _thisControl.transform.position;

                    if (_movie_start_event != null) _movie_start_event.Trigger();
                }
                else
                {
                    //_thisControl.transform.position = _base_vector;
                    if (_movie_end_event != null) _movie_end_event.Trigger();
                }
            }
            
        }

        private void OnVibrationWidthChanged(JSONStorableFloat val)
        {
            _move_width = val.val * 0.01f;
            SetVibrationValue();
        }

        private void OnIntervalTimeChanged(JSONStorableFloat val)
        {
            _move_time = val.val;
            SetVibrationValue();
        }

        private void SetVibrationValue()
        {
            _move_gap = _move_width / _move_time;
        }

        private void OnDirectionChanged(JSONStorableFloat val)
        {

        }

        private void OnMovieStartEventFunc()
        {
            if(_movie_start_event == null) _movie_start_event = new EventTrigger(this, "Movie Start Event");
            _movie_start_event.OpenPanelActionStart();
        }

        private void OnMovieEndEventFunc()
        {
            if (_movie_end_event == null) _movie_end_event = new EventTrigger(this, "Movie End Event");
            _movie_end_event.OpenPanelActionStart();
        }

    }

    public class EventTrigger : Trigger
    {
        public MVRScript Owner
        {
            get; private set;
        }

        private string _name;

        public string Name
        {
            get { return _name; }
        }

        public EventTrigger(MVRScript owner, string name)
        {
            Owner = owner;
            _name = name;
            handler = DMEventHandler.Instance;
        }

        public void OpenPanelActionStart()
        {
            if (!DMEventHandler.Loaded)
            {
                SuperController.LogError("EventTrigger: You need to call DMEventHandler.LoadAssets() before use.");
                return;
            }

            triggerActionsParent = Owner.UITransform;
            InitTriggerUI();
            OpenTriggerActionsPanel();

            triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = _name;
            triggerActionsPanel.Find("Content/Tab1/Label").GetComponent<Text>().text = "Event Actions";
            triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(false);
            triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
        }

        public void RestoreFromJSON(JSONClass jc, string subScenePrefix, bool isMerge, bool setMissingToDefault)
        {
            if (jc.HasKey(_name))
            {
                JSONClass tc = jc[_name].AsObject;
                if (tc != null)
                    base.RestoreFromJSON(tc, subScenePrefix, isMerge);
            }
            else if (setMissingToDefault)
            {
                base.RestoreFromJSON(new JSONClass());
            }
        }

        public void Trigger()
        {
            active = true;
            active = false;
        }
    }

    public class DMEventHandler : TriggerHandler
    {
        public static bool Loaded { get; private set; }

        private static DMEventHandler myInstance;

        private RectTransform myTriggerActionsPrefab;
        private RectTransform myTriggerActionMiniPrefab;
        private RectTransform myTriggerActionDiscretePrefab;
        private RectTransform myTriggerActionTransitionPrefab;

        public static DMEventHandler Instance
        {
            get
            {
                if (myInstance == null)
                    myInstance = new DMEventHandler();
                return myInstance;
            }
        }

        public static void LoadAssets()
        {
            SuperController.singleton.StartCoroutine(Instance.LoadAssetsInternal());
        }

        private IEnumerator LoadAssetsInternal()
        {
            foreach (var x in LoadAsset("z_ui2", "TriggerActionsPanel", p => myTriggerActionsPrefab = p))
                yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionMiniPanel", p => myTriggerActionMiniPrefab = p))
                yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionDiscretePanel", p => myTriggerActionDiscretePrefab = p))
                yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionTransitionPanel", p => myTriggerActionTransitionPrefab = p))
                yield return x;

            Loaded = true;
        }

        private IEnumerable LoadAsset(string assetBundleName, string assetName, Action<RectTransform> assign)
        {
            AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync(assetBundleName, assetName, typeof(GameObject));
            if (request == null)
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null request.");
            yield return request;
            GameObject go = request.GetAsset<GameObject>();
            if (go == null)
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null GameObject.");
            RectTransform prefab = go.GetComponent<RectTransform>();
            if (prefab == null)
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null RectTansform.");
            assign(prefab);
        }


        void TriggerHandler.RemoveTrigger(Trigger t)
        {
            // nothing to do
        }

        void TriggerHandler.DuplicateTrigger(Trigger t)
        {
            throw new NotImplementedException();
        }

        RectTransform TriggerHandler.CreateTriggerActionsUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionsPrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionMiniUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionMiniPrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionDiscreteUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionDiscretePrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionTransitionUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionTransitionPrefab);
        }

        void TriggerHandler.RemoveTriggerActionUI(RectTransform rt)
        {
            UnityEngine.Object.Destroy(rt?.gameObject);
        }
    }

}